# coding: utf-8

# -------------------------------------------------------------------
# Enveloppe convexe
# Marc Lorenzi
# 4 janvier 2019
# -------------------------------------------------------------------

import random
from math import sqrt
from tkinter import *
import time

# -------------------------------------------------------------------
# Variables globales

NPTS = 50 # Nombre de points de l'ensemble
SLP = 0.5  # Délai en secondes après chaque tracé de segment 
W = 600    # Largeur du Canvas
H = 600    # Hauteur du Canvas
# -------------------------------------------------------------------


# -------------------------------------------------------------------
# Loi binomiale bivariée

def bivariate(mu1, mu2, s1, s2, r):
    z1 = random.gauss(0, 1)
    z2 = random.gauss(0, 1)
    x1 = mu1 + s1 * z1
    x2 = mu2 + s2 * (z1 * r + z2 * sqrt(1 - r ** 2))
    return (x1, x2)

# Liste de points aléatoires

def random_list(n):
    M = 1000
    s = []
    cnt = 0
    for k in range(n):
        x, y = bivariate(0, 0, 1000, 1000, 0.3)
        x, y = int(x), int(y)
        if not (x, y) in s: s.append((x, y))
    return s

# -------------------------------------------------------------------
# Points et vecteurs


# Déterminant

def determinant(u, v):
    return u[0] * v[1] - u[1] * v[0]

# Vecteur

def vecteur(A, B):
    return (B[0] - A[0], B[1] - A[1])


# Famille liée, libre directe, libre indirecte

def pseudo_angle(A, B, C):
    u = vecteur(A, B)
    v = vecteur(A, C)
    return determinant(u, v)

# distance au carré

def distance(A, B):
    x, y = vecteur(A, B)
    return x ** 2 + y ** 2 

# -------------------------------------------------------------------
# Ordre lexicographique

def ordre_lexico(A, B):
    return (A[1] < B[1]) or (A[1] == B[1] and A[0] <= B[0])


# Point minimal de la liste pour l'ordre lexicographique

def min_lexico(s):
    m = s[0]
    for x in s:
        if ordre_lexico(x, m): m = x
    return m

# -------------------------------------------------------------------
# Ordre angle-distance

def comp(Omega, A, B):
    t = pseudo_angle(Omega, A, B)
    if t > 0: return True
    elif t < 0: return False
    else: return distance(Omega, A) <= distance(Omega, B)

# -------------------------------------------------------------------
# Trier par angles et distances

def partition(s, l, r, ordre):
    k = random.randint(l, r)
    s[k], s[r] = s[r], s[k]
    i = l - 1
    for j in range(l, r):
        if ordre(s[j], s[r]):
            i = i + 1
            s[i], s[j] = s[j], s[i]
    s[i + 1], s[r] = s[r], s[i + 1]
    return i + 1

def tri_aux(s, l, r, ordre):
    if l <= r:
        q = partition(s, l, r, ordre)
        tri_aux(s, l, q - 1, ordre)
        tri_aux(s, q + 1, r, ordre)
        
def quicksort(s, ordre):
    tri_aux(s, 0, len(s) - 1, ordre)


def trier_angles_distances(Omega, s):
    ordre = lambda A, B: comp(Omega, A, B)
    quicksort(s, ordre)

# -------------------------------------------------------------------
# Opérations de pile

def pile_vide(): return []
def empiler(S, A): S.append(A)
def depiler(S): S.pop()
    
def sommet(S): return S[-1]
def sous_sommet(S): return S[-2]

# -------------------------------------------------------------------
# Supprimer alignements avec p_0

def supprimer_alignes(s, Omega):
    s1 = [s[0]]
    n = len(s)
    k = 1
    while k < n:
        k = suppr_aux(s, k, Omega)
        s1.append(s[k - 1])
    return s1

def suppr_aux(s, k, Omega):
    p = s[k]
    j = k + 1
    while j < len(s) and pseudo_angle(Omega, p, s[j]) == 0: j = j + 1
    return j

# -------------------------------------------------------------------
# Animation de l'algorithme de Graham

def anim_graham(A, c):
    lines = {}
    points = {}
    bounds = bornes(A)
    r = 4
    S = pile_vide()
    Omega = min_lexico(A)
    trier_angles_distances(Omega, A)
    A1 = supprimer_alignes(A, Omega)
    for i in range(len(A1)):
        u, v = scale(A1[i], bounds)
        points[i] = c.create_oval([(u - r, v - r), (u + r, v + r)], fill='white')
    empiler(S, A1[0])
    empiler(S, A1[1])
    empiler(S, A1[2])
    c.itemconfig(points[0], fill='orange')
    c.itemconfig(points[1], fill='orange')
    c.itemconfig(points[2], fill='orange')
    creer_segment(S[0], S[1], bounds, lines, c)
    creer_segment(S[1], S[2], bounds, lines, c)
    c.itemconfig(lines[(S[0], S[1])], fill='red', width=2)
    c.itemconfig(lines[(S[1], S[2])], fill='red', width=2)
    for i in range(3, len(A1)):
        c.itemconfig(points[i], fill='orange')
        creer_segment(sommet(S), A1[i], bounds, lines, c)
        while pseudo_angle(sous_sommet(S), sommet(S), A1[i]) <= 0:
            p = sommet(S)
            q = sous_sommet(S)
            creer_segment(q, A1[i], bounds, lines, c)
            supprimer_segment(p, A1[i], lines, c)
            supprimer_segment(q, p, lines, c)
            depiler(S)
        p = sommet(S)
        q = sous_sommet(S)
        c.itemconfig(lines[(p, A1[i])], fill='red', width=2)
        c.update()
        empiler(S, A1[i])
    
    creer_segment(S[-1], S[0], bounds, lines, c)
    c.itemconfig(lines[(S[-1], S[0])], fill='red', width=2)

# -------------------------------------------------------------------
# Bornes d'un ensemble de points

def bornes(A):
    x, y = A[0]
    xmin, xmax, ymin, ymax = x, x, y, y
    for x, y in A:
        if x < xmin: xmin = x
        if x > xmax: xmax = x
        if y < ymin: ymin = y
        if y > ymax: ymax = y
    return (xmin, xmax, ymin, ymax)

# -------------------------------------------------------------------
# Mettre un point à l'échelle du Canvas

def scale(p, bounds):
    x, y = p
    xmin, xmax, ymin, ymax = bounds
    u = 20 + ((W - 40) * (x - xmin)) // (xmax - xmin)
    v = 20 + ((H - 40) * (y - ymax)) // (ymin - ymax)
    return (u, v) 

# Créer et supprimer des segments sur le Canvas

def creer_segment(A, B, bounds, lines, c):
    a, b = scale(A, bounds)
    u, v = scale(B, bounds)
    lines[(A, B)] = c.create_line([(a, b), (u, v)], fill='green')
    c.update()
    time.sleep(SLP)

def supprimer_segment(A, B, lines, c):
    c.delete(lines[(A, B)])
    del lines[(A, B)]
    c.update()
    time.sleep(SLP)

# -------------------------------------------------------------------
# Main

if __name__ == '__main__':
    root = Tk()
    root.title('Graham')
    c = Canvas(root, width=W, height=H, bg='black')
    c.pack()
    A = random_list(NPTS)
    bounds = bornes(A)
    anim_graham(A, c)
    root.mainloop()
